// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    instruction::{PRegNum, CONST_ZERO_UNIT_NUM},
    unit::UnitKind,
};
use fayalite::{
    intern::{Intern, Interned},
    prelude::*,
};
use serde::{Deserialize, Serialize};
use std::num::NonZeroUsize;

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug, Serialize, Deserialize)]
#[non_exhaustive]
pub struct UnitConfig {
    pub kind: UnitKind,
    /// max number of instructions that can be in-flight through this unit at any point in time.
    pub max_in_flight: Option<NonZeroUsize>,
}

impl UnitConfig {
    pub fn new(kind: UnitKind) -> Self {
        Self {
            kind,
            max_in_flight: None,
        }
    }
}

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug, Serialize, Deserialize)]
#[non_exhaustive]
pub struct CpuConfig {
    pub units: Interned<[UnitConfig]>,
    pub out_reg_num_width: usize,
    pub fetch_width: NonZeroUsize,
    /// default value for [`UnitConfig::max_in_flight`]
    pub default_unit_max_in_flight: NonZeroUsize,
}

impl CpuConfig {
    pub const DEFAULT_OUT_REG_NUM_WIDTH: usize = 4;
    pub const DEFAULT_FETCH_WIDTH: NonZeroUsize = {
        let Some(v) = NonZeroUsize::new(1) else {
            unreachable!();
        };
        v
    };
    pub const DEFAULT_UNIT_MAX_IN_FLIGHT: NonZeroUsize = {
        let Some(v) = NonZeroUsize::new(8) else {
            unreachable!();
        };
        v
    };
    pub fn new(units: Interned<[UnitConfig]>) -> Self {
        Self {
            units,
            out_reg_num_width: Self::DEFAULT_OUT_REG_NUM_WIDTH,
            fetch_width: Self::DEFAULT_FETCH_WIDTH,
            default_unit_max_in_flight: Self::DEFAULT_UNIT_MAX_IN_FLIGHT,
        }
    }
    pub fn non_const_unit_nums(&self) -> std::ops::Range<usize> {
        (CONST_ZERO_UNIT_NUM + 1)..(self.units.len() + 1)
    }
    pub fn unit_num_width(&self) -> usize {
        UInt::range(CONST_ZERO_UNIT_NUM..self.non_const_unit_nums().end).width()
    }
    pub fn unit_max_in_flight(&self, unit_index: usize) -> NonZeroUsize {
        self.units[unit_index]
            .max_in_flight
            .unwrap_or(self.default_unit_max_in_flight)
    }
    pub fn retire_queue_index_width(&self) -> usize {
        let max_in_flight: usize = (0..self.units.len())
            .map(|unit_index| self.unit_max_in_flight(unit_index).get())
            .sum();
        2 + max_in_flight.next_power_of_two().ilog2() as usize
    }
}

mod sealed {
    pub trait Sealed {}
}

impl sealed::Sealed for PhantomConst<CpuConfig> {}

pub trait CpuConfigType: Type + ToExpr<Type = Self> + sealed::Sealed {
    fn get(self) -> Interned<CpuConfig>;
}

impl CpuConfigType for PhantomConst<CpuConfig> {
    fn get(self) -> Interned<CpuConfig> {
        self.get()
    }
}

pub trait Identity<Arg: ?Sized> {
    type SelfType: ?Sized;
    type ArgType: ?Sized;
}

impl<T: ?Sized, Arg: ?Sized> Identity<Arg> for T {
    type SelfType = T;
    type ArgType = Arg;
}

macro_rules! impl_cpu_config_accessors {
    (
        $(
            #[without_generics = $without_generics:ident]
            $vis:vis type $ident:ident<$T:ident> = |$arg:ident| $expr:expr;
        )*
    ) => {
        $(
            #[allow(non_camel_case_types)]
            $vis struct $without_generics;

            #[allow(non_upper_case_globals)]
            $vis const $ident: $without_generics = $without_generics;

            $vis type $ident<$T> = <DynSize as Identity<$T>>::SelfType;

            impl<$T: CpuConfigType> std::ops::Index<$T> for $without_generics {
                type Output = usize;

                fn index(&self, $arg: $T) -> &Self::Output {
                    Interned::into_inner(Intern::intern_sized($expr))
                }
            }
        )*
    };
}

impl_cpu_config_accessors! {
    #[without_generics = __UnitNumWidth_WithoutGenerics]
    pub type UnitNumWidth<T> = |arg| arg.get().unit_num_width();
    #[without_generics = __UnitOutRegNumWidth_WithoutGenerics]
    pub type UnitOutRegNumWidth<T> = |arg| arg.get().out_reg_num_width;
    #[without_generics = __PRegNumWidth_WithoutGenerics]
    pub type PRegNumWidth<T> = |arg| PRegNum[arg].canonical().bit_width();
    #[without_generics = __RetireQueueIndexWidth_WithoutGenerics]
    pub type RetireQueueIndexWidth<T> = |arg| arg.get().retire_queue_index_width();
    #[without_generics = __UnitCount_WithoutGenerics]
    pub type UnitCount<T> = |arg| arg.get().non_const_unit_nums().len();
    #[without_generics = __FetchWidth_WithoutGenerics]
    pub type FetchWidth<T> = |arg| arg.get().fetch_width.get();
}
