// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    config::{CpuConfig, CpuConfigType, FetchWidth, PRegNumWidth},
    instruction::{MOp, MOpDestReg, MOpRegNum, MOpTrait, MoveRegMOp, PRegNum, RenamedMOp},
    rename_table::{rename_table, RenameTablePortConfig},
    unit::{RenamedInsnData, RetireQueueIndex, UnitMOp},
    util::array_vec::{ArrayVec, Length, ReadyValidArray},
};
use fayalite::{
    prelude::*,
    util::{prefix_sum::PrefixSumAlgorithm, ready_valid::ReadyValid},
};

#[hdl(no_static)]
pub struct InstructionRenameInputInsn<C: Type + CpuConfigType> {
    pub mop: MOp,
    pub pc: UInt<64>,
    pub renamed_dest: PRegNum<C>,
}

#[hdl(no_static)]
struct InsnsInPrefixSummary<C: Type + CpuConfigType> {
    all_ready: Bool,
    ready_count: Length<FetchWidth<C>>,
    retire_queue_used: Length<FetchWidth<C>>,
    config: C,
}

type C = PhantomConst<CpuConfig>;

#[hdl]
pub type InstructionRenameInsnsOut<C: Type + CpuConfigType> = ArrayType<
    ReadyValid<RenamedInsnData<C, RenamedMOp<PRegNumWidth<C>>, PRegNum<C>>>,
    FetchWidth<C>,
>;

#[hdl_module]
pub fn instruction_rename(config: PhantomConst<CpuConfig>) {
    #[hdl]
    let cd: ClockDomain = m.input();
    #[hdl]
    let insns_in: ReadyValidArray<InstructionRenameInputInsn<C>, FetchWidth<C>> =
        m.input(ReadyValidArray[InstructionRenameInputInsn[config]][FetchWidth[config]]);
    #[hdl]
    let start_retire_queue_index: RetireQueueIndex<C> = m.input(RetireQueueIndex[config]);
    #[hdl]
    let end_retire_queue_index: RetireQueueIndex<C> = m.output(RetireQueueIndex[config]);
    #[hdl]
    let insns_out: InstructionRenameInsnsOut<C> = m.output(InstructionRenameInsnsOut[config]);

    // TODO: handle resetting table after cancelling instructions

    #[hdl]
    let insns_ready_or_move = wire(ArrayType[Bool][FetchWidth[config]]);

    for (insn_ready_or_move, insn_out) in insns_ready_or_move.into_iter().zip(insns_out) {
        connect(insn_ready_or_move, insn_out.ready);
    }

    ArrayVec::for_each(insns_in.data, |fetch_index, input_insn| {
        #[hdl]
        match input_insn.mop {
            UnitMOp::<_, _, _>::TransformedMove(_) => {
                connect(insns_ready_or_move[fetch_index], true);
            }
            UnitMOp::<_, _, _>::AluBranch(_) | UnitMOp::<_, _, _>::LoadStore(_) => {}
        }
    });

    let insns_in_prefix_summary_ty = InsnsInPrefixSummary[config];
    #[hdl]
    let insns_in_prefix_summaries = wire(ArrayType[insns_in_prefix_summary_ty][FetchWidth[config]]);
    let insns_in_prefix_summaries_vec = PrefixSumAlgorithm::WorkEfficient.run(
        (0..FetchWidth[config]).map(|fetch_index| {
            #[hdl]
            let insns_in_prefix_summary_in = wire(insns_in_prefix_summary_ty);
            #[hdl]
            let InsnsInPrefixSummary::<_> {
                all_ready,
                ready_count,
                retire_queue_used,
                config: _,
            } = insns_in_prefix_summary_in;
            connect(all_ready, insns_out[fetch_index].ready);
            connect(
                ready_count,
                Expr::ty(ready_count).cast_from_uint_unchecked(all_ready.cast_to(UInt[1])),
            );
            connect(retire_queue_used, Expr::ty(retire_queue_used).zero());
            #[hdl]
            if let HdlSome(input_insn) = ArrayVec::get(insns_in.data, fetch_index) {
                connect(retire_queue_used, ready_count);
                #[hdl]
                match input_insn.mop {
                    UnitMOp::<_, _, _>::TransformedMove(_) => {
                        connect(all_ready, true);
                    }
                    UnitMOp::<_, _, _>::AluBranch(_) | UnitMOp::<_, _, _>::LoadStore(_) => {}
                }
            }
            insns_in_prefix_summary_in
        }),
        |l, r| {
            #[hdl]
            let insns_in_prefix_summary_merge = wire(insns_in_prefix_summary_ty);
            #[hdl]
            let InsnsInPrefixSummary::<_> {
                all_ready,
                ready_count,
                retire_queue_used,
                config: _,
            } = insns_in_prefix_summary_merge;
            connect(all_ready, l.all_ready & r.all_ready);
            #[hdl]
            if l.all_ready {
                connect(
                    ready_count,
                    Expr::ty(ready_count).cast_from_uint_unchecked(
                        Length::as_uint(l.ready_count) + Length::as_uint(r.ready_count),
                    ),
                );
                connect(
                    retire_queue_used,
                    Expr::ty(retire_queue_used).cast_from_uint_unchecked(
                        Length::as_uint(l.retire_queue_used) + Length::as_uint(r.retire_queue_used),
                    ),
                );
            } else {
                connect(ready_count, l.ready_count);
                connect(retire_queue_used, l.retire_queue_used);
            }
            insns_in_prefix_summary_merge
        },
    );
    for (l, r) in insns_in_prefix_summaries
        .into_iter()
        .zip(insns_in_prefix_summaries_vec)
    {
        connect(l, r);
    }
    connect(
        insns_in.ready,
        insns_in_prefix_summaries[FetchWidth[config] - 1].ready_count,
    );

    #[hdl]
    let retire_queue_indexes = wire(Array[RetireQueueIndex[config]][FetchWidth[config] + 1]);
    connect(retire_queue_indexes[0], start_retire_queue_index);
    connect(
        end_retire_queue_index,
        retire_queue_indexes[FetchWidth[config]],
    );
    for (retire_queue_index, insns_in_prefix_summary) in retire_queue_indexes
        .into_iter()
        .skip(1)
        .zip(insns_in_prefix_summaries)
    {
        connect_any(
            retire_queue_index.index,
            start_retire_queue_index.index
                + Length::as_uint(insns_in_prefix_summary.retire_queue_used),
        );
    }

    let mut port_configs = Vec::new();
    let mut src_reg_count = 0;
    MOpTrait::for_each_src_reg(MOp.uninit(), &mut |_, src_index| {
        src_reg_count = src_reg_count.max(src_index + 1);
    });
    for _ in 0..FetchWidth[config] {
        for _ in 0..src_reg_count {
            port_configs.push(RenameTablePortConfig::Read {
                addr_range: MOpRegNum::NON_CONST_REG_NUMS,
            });
        }
        for dest_reg_kind in MOpDestReg::REG_KINDS {
            port_configs.push(RenameTablePortConfig::Write {
                addr_range: dest_reg_kind.reg_num_range(),
            });
        }
    }

    #[hdl]
    let rename_table = instance(rename_table(config, &port_configs));

    connect(rename_table.cd, cd);

    for read_port in rename_table.read_ports {
        connect_any(read_port.addr, 0_hdl_u0);
    }
    for write_port in rename_table.write_ports {
        connect_any(write_port.addr, 0_hdl_u0);
        connect_any(write_port.data, PRegNum[config].const_zero());
    }

    ArrayVec::for_each(
        ReadyValidArray::firing_data(insns_in),
        |fetch_index, input_insn| {
            let read_port_index = fetch_index * src_reg_count;
            let write_port_index = fetch_index * MOpDestReg::REG_COUNT;
            #[hdl]
            let InstructionRenameInputInsn::<_> {
                mop,
                pc,
                renamed_dest,
            } = input_insn;
            let insn_out =
                MOpTrait::map_regs(mop, (), PRegNumWidth[config], &mut |src_reg, src_index| {
                    connect(
                        rename_table.read_ports[read_port_index + src_index].addr,
                        src_reg.cast_bits_to(MOpRegNum),
                    );
                    rename_table.read_ports[read_port_index + src_index]
                        .data
                        .cast_to_bits()
                });
            for (i, dest_reg) in MOpDestReg::regs(MOpTrait::dest_reg(mop))
                .into_iter()
                .enumerate()
            {
                connect(
                    rename_table.write_ports[write_port_index + i].addr,
                    dest_reg,
                );
                connect(
                    rename_table.write_ports[write_port_index + i].data,
                    renamed_dest,
                );
            }
            let insn_out = UnitMOp::try_with_transformed_move_op(
                insn_out,
                RenamedMOp[PRegNumWidth[config]].TransformedMove,
                |insn_out: Expr<HdlOption<_>>, move_reg: Expr<MoveRegMOp<_, _>>| {
                    for i in 0..MOpDestReg::REG_COUNT {
                        // execute move by using same PRegNum as src[0] for dest
                        connect(
                            rename_table.write_ports[write_port_index + i].data,
                            move_reg.common.src[0].cast_bits_to(PRegNum[config]),
                        );
                    }
                    // move already executed, so remove it
                    connect(insn_out, Expr::ty(insn_out).HdlNone());
                },
            );
            connect(
                insns_out[fetch_index].data,
                HdlOption::map(insn_out, |insn_out| {
                    #[hdl]
                    RenamedInsnData::<_, _, _> {
                        retire_queue_index: retire_queue_indexes[fetch_index],
                        pc,
                        dest: renamed_dest,
                        mop: insn_out,
                    }
                }),
            );
        },
    );
}
